(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2020 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`Relaxation`", {"ProcessOptions`", "AtomicDensityMatrix`", "AtomicDensityMatrix`WignerRotate`", "AtomicDensityMatrix`TensorAlgebra`"}];

unitSphericalVector::usage = "";

ikk::usage = "";

Begin["`Private`"];

IntrinsicRelaxation[sys_?AtomicSystemQ] = WignerEckart[sys, {NaturalWidth, 0, 0}];

TransitRelaxation[sys_?AtomicSystemQ, gamma_] := gamma WignerEckart[sys, {Identity, 0, 0}];

Options[TransitRepopulation] = FilterRules[Options@DMVariables, Except@DMLabel];

(TransitRepopulation[as_?AtomicSystemQ, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    (TransitRepopulation[as, Density@# TransitRate[VoidRegion][#]] +
        Sum[
            Density@# / Density@l TransitRate[l][#] *
                DensityMatrix[as, DMLabel -> RegionLabel@l, FilterRules@opts],
            {l, es}
        ]
    )& /@ es
) // ProcessOptions;

TransitRepopulation[sys_?AtomicSystemQ, gamma_] := Module[
    {g = SublevelMultiplicity@GroundStates@sys},
    gamma WignerEckart[sys, {GroundState, 0, 0}] / If[g == 0, 1, g]
];

Options[OpticalRepopulation] = Options@DMVariables;

OptionValidator[ExpandDipoleRME] = True | False;

(OpticalRepopulation[sys_?AtomicSystemQ, opts : OptionsPattern[]] :=
    Module[{representation},
        Switch[representation,
            {"Zeeman"},
            OpticalRepopulationZ[sys, FilterRules@opts],
            {"PolarizationMoments", ___},
            OpticalRepopulationPM[sys, Rest@representation, FilterRules@opts]
        ]
    ]) // ProcessOptions;

Options[OpticalRepopulationZ] = Options@OpticalRepopulation;

branchingRatioMatrix[sys_] := SparseArray@Sqrt[
    (2 J[sys] + 1) NaturalWidth[sys] DirectSelfProduct[BranchingRatio[#2][#1] &, sys]
];

energyDifferenceMatrix[sys_] := SparseArray@Sqrt@DirectSelfProduct[Subtract, Energy@sys];

degeneracyMatrix[sys_] :=
    SparseArray@DirectSelfProduct[
        Composition[Boole, PossibleZeroQ, Subtract],
        Energy@sys + HyperfineShift@sys
    ];

(OpticalRepopulationZ[sys_?AtomicSystemQ, opts : OptionsPattern[]] :=
    Module[{d, branchingMat, degenerateStates, dm, rank},
        (*spherical components of the electric dipole operator*)
        d = SparseArray@WignerEckart[sys, {"NormalizedDipole", 1}];
        (* Matrix of branching rates *)
        branchingMat = branchingRatioMatrix[sys];
        (* Only allow transfer of coherences from a pair of degenerate levels to another pair of degenerate levels *)
        degenerateStates = degeneracyMatrix[sys];
        dm = DensityMatrix[sys, FilterRules@opts] degenerateStates;
        rank = Rank@d;
        Normal[degenerateStates *
            Sum[
                (-1)^q (Transpose@branchingMat Component[d, q]).dm.(branchingMat Component[d, -q]),
                {q, -rank, rank}
            ]
        ]
    ]) // ProcessOptions;

Options[OpticalRepopulationPM] = Options@PMVariables;

(OpticalRepopulationPM[sys_?AtomicSystemQ, opts : OptionsPattern[]] /; AtomicSystemType[sys] == "Toy" :=
    Map[{{{#}}}&, OpticalRepopulationZ[sys, Representation -> "Zeeman", FilterRules@opts], {2}]
) // ProcessOptions;

(OpticalRepopulationPM[sys_?AtomicSystemQ, opts : OptionsPattern[]] :=
    Module[{levels, excitedLevels, pme, ne, n, fe, f, repop, re, r, rate, hf},
        levels = RemoveZeemanSublevels@sys;
        excitedLevels = ExcitedStates@levels;
        pme = PolarizationMoments[excitedLevels, FilterRules@opts];
        repop = ZeroTensorDecompositionMatrix[sys, FilterRules@opts];
        {n, ne} = Length /@ {levels, excitedLevels};
        {f, fe} = F /@ {levels, excitedLevels};
        {r, re} = Diagonal /@ Rank /@ {repop, pme};
        rate = Table[(2 J@le + 1) NaturalWidth@le BranchingRatio[l][le], {l, levels}, {le, excitedLevels}];
        hf = If[AtomicSystemType@sys === "HyperfineZeeman",
            Table[hyperfineRMEFactor[l, 1, le] hyperfineRMEFactor[le, 1, l], {l, levels}, {le, excitedLevels}],
            ConstantArray[1, {Length@levels, Length@excitedLevels}]
        ];
        Do[
            repop[[i, i, k - r[[i, 1]] + 1]] +=
                (-1)^(2 f[[i]] + k) hf[[i, j]] rate[[i, j]] Component[pme[[j, j]], k] SJS[{k, f[[i]], f[[i]]}, {1, fe[[j]], fe[[j]]}],
            {i, n},
            {j, ne},
            {k, Max[r[[i, 1]], re[[j, 1]]], Min[r[[i, -1]], re[[j, -1]]]}
        ];
        repop
    ]) // ProcessOptions;

OpticalRepopulation[as_?AtomicSystemQ, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    Module[
        {label, repopulation},
        repopulation = EvaluatedFunction[label, OpticalRepopulation[as, DMLabel -> label, opts]];
        Table[repopulation@RegionLabel@j, {j, es}]
    ];

OpticalRepopulationOperator[sys_?AtomicSystemQ] :=
    Module[{dm, or},
        dm = DensityMatrix[sys, ComplexExpandVariables -> False, Representation -> "Zeeman"];
        or = OpticalRepopulation[sys, ComplexExpandVariables -> False];
        Normal@Transpose@Map[Partition[#, Length[sys]]&, Last@CoefficientArrays[or, Flatten@dm], {2}]
    ];

(*The OpticalRepopulation matrix is equal to Map[Tr[#\[CenterDot]DensityMatrix[System]]&,OpticalRepopulationOperator[System],{2}]*)

(RecoilRepopulation[sys_?AtomicSystemQ, velocityGroupRecoilMomentum_, opts : OptionsPattern[]] :=
    Module[{d, branchingMat, degenerateStates, dm, rank},
        If[PossibleZeroQ@velocityGroupRecoilMomentum, Return[WignerEckart[sys, {Zero, 0, 0}, FilterRules@opts]]];
        (*spherical components of the electric dipole operator*)
        d = SparseArray@WignerEckart[sys, {"NormalizedDipole", 1}];
        (* Matrix of branching rates *)
        branchingMat = SparseArray@Sqrt[(2 J@sys + 1) NaturalWidth@sys DirectSelfProduct[Wavenumber[StateLabel@#1 -> StateLabel@#2] BranchingRatio[#2][#1] &, sys]];
        (* Only allow transfer of coherences from a pair of degenerate levels to another pair of degenerate levels *)
        degenerateStates = degeneracyMatrix[sys];
        dm = DensityMatrix[sys, FilterRules@opts] degenerateStates;
        rank = Rank@d;
        Normal[degenerateStates *
            Sum[
                (-1)^q (Transpose@branchingMat Component[d, q]).dm.(branchingMat Component[d, -q]),
                {q, -rank, rank}
            ]
        ] / velocityGroupRecoilMomentum
    ]) // ProcessOptions;

(RecoilRepopulation[as_?AtomicSystemQ, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    Module[
        {label, repopulation, momentum, zero},
        repopulation = EvaluatedFunction[{label, momentum}, RecoilRepopulation[as, momentum, DMLabel -> label, opts]];
        zero = WignerEckart[as, {Zero, 0, 0}, FilterRules@opts];
        Table[
            Sum[
                Which[
                    i === j,
                    -Sum[If[DefinedQ[i, VelocityGroupRecoilMomentum@RegionLabel@k], repopulation[RegionLabel@i, VelocityGroupRecoilMomentum[k][i]], zero], {k, es}],
                    DefinedQ[i, VelocityGroupRecoilMomentum@RegionLabel@j],
                    Density@j / Density@i repopulation[RegionLabel@i, VelocityGroupRecoilMomentum[j][i]],
                    True,
                    0
                ],
                {i, es}
            ],
            {j, es}
        ]
    ]) // ProcessOptions;

Options[SDamping] = Join[Options[DMVariables], Thread[Options[WignerEckart][[All, 1]] -> Automatic]];

SDamping[atomicSystem_?AtomicSystemQ, experimentalSystem : {_ExperimentalRegion..}, opts: OptionsPattern[]] /;
    AtomicSystemType[atomicSystem] == "HyperfineZeeman" := Module[
    {sDamping, label, rate, rates},
    rates = SDampingRate[experimentalSystem] Boole[DefinedQ[#, SDampingRate]& /@ experimentalSystem];
    If[MatchQ[rates, {0 ..}],
        ConstantArray[WignerEckart[atomicSystem, {Zero, 0, 0}], Length@experimentalSystem],
        sDamping = EvaluatedFunction[{label, rate}, SDamping[atomicSystem, rate, DMLabel -> label, opts]];
        MapThread[sDamping, {RegionLabel@experimentalSystem, rates}]
    ]
];

SDamping[atomicSystem_?AtomicSystemQ, experimentalSystem : {_ExperimentalRegion..}, opts: OptionsPattern[]] :=
    ConstantArray[WignerEckart[atomicSystem, {Zero, 0, 0}], Length@experimentalSystem];

(SDamping[sys_?AtomicSystemQ, rate_, opts : OptionsPattern[]] /;
    AtomicSystemType[sys] == "HyperfineZeeman" := Module[{allowedElements, allowedCouplings, dm, s},
    dm = DensityMatrix[sys, AllowedElements -> GroundStateQ && GroundStateQ' && allowedElements, FilterRules@opts];
    s = ToCartesian@WignerEckart[sys, {J, 1}, AllowedCouplings -> GroundStateQ && GroundStateQ' && allowedCouplings, FilterRules@opts];
    -rate Expand[CenterDot[s, s, dm] - CenterDot[s, dm, s]]
]) // ProcessOptions;

SDamping[sys_?AtomicSystemQ, rate_, opts : OptionsPattern[]] := WignerEckart[sys, {Zero, 0, 0}];

Options[OpticalBroadening] = Options@DMVariables;

(OpticalBroadening[sys_?AtomicSystemQ, transition_, width_, opts : OptionsPattern[]] := Module[
    {allowedElements},
    -(width/2) DensityMatrix[sys, AllowedElements -> transition && Parity != Parity' && StateLabel != StateLabel' && allowedElements, FilterRules@opts]
]) // ProcessOptions;

OpticalBroadening[sys_?AtomicSystemQ, region_ExperimentalRegion, opts : OptionsPattern[]] := Module[
    {opticalParameters, transitions, opticalWidths},
    opticalParameters = OptionProcessor[OpticalParameters]@OpticalParameters[region];
    {transitions, opticalWidths} = Transpose[{AllowedCouplings, OpticalWidth} /. Last /@ opticalParameters];
    Total@MapThread[OpticalBroadening[sys, #1, #2, DMLabel -> RegionLabel@region, opts]&, {transitions, opticalWidths}]
];

OpticalBroadening[sys_?AtomicSystemQ, expSys_?ExperimentalSystemQ, opts : OptionsPattern[]] :=
    OpticalBroadening[sys, #, opts]& /@ expSys;

Relaxation[as_?AtomicSystemQ, es : {_ExperimentalRegion..}] := (
    IntrinsicRelaxation[as] +
        TransitRelaxation[as, TransitRate[VoidRegion][#]] +
            TransitRelaxation[as, Sum[TransitRate[l][#], {l, es}]]
)& /@ es;

Options[Repopulation] = FilterRules[Options@DMVariables, Except@DMLabel];

(Repopulation[as_?AtomicSystemQ, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    TransitRepopulation[as, es, FilterRules@opts] +
        OpticalRepopulation[as, es, FilterRules@opts] +
            If[DefinedQ[es, VelocityGroupRecoilMomentum], RecoilRepopulation[as, es, FilterRules@opts], 0] +
                SDamping[as, es, FilterRules@opts]
) // ProcessOptions;

FluorescenceOperator[sys_?AtomicSystemQ, freq_, th_, phi_] :=
    Module[{dc, gr, ikk},
        If[AtomicSystemType[sys] == "Toy", Return[FluorescenceOperator[sys, freq] / (4 Pi)]];
        dc = ExpandDipoleRME[sys, ToCartesian@WignerEckart[sys, {Dipole, 1}]];
        gr = DiagonalMatrix[Boole /@ GroundStateQ /@ sys];
        ikk = IdentityMatrix@3 - DirectSelfProduct[{Cos[phi] Sin[th], Sin[phi] Sin[th], Cos[th]}];
        freq^4 / (2\[Pi]) Sum[dc[[i]].gr.dc[[j]] ikk[[i, j]], {i, 3}, {j, 3}]
    ];

FluorescenceOperator[sys_?AtomicSystemQ, freq_] :=
    Module[{dc, gr},
        dc = ExpandDipoleRME[sys, ToCartesian@WignerEckart[sys, {Dipole, 1}]];
        gr = DiagonalMatrix[Boole /@ GroundStateQ /@ sys];
        4 / 3 freq^4 Sum[dc[[i]].gr.dc[[i]], {i, Length@dc}]
    ];

Options[Fluorescence] = Options@DMVariables;

Fluorescence[sys_?AtomicSystemQ, upperState_ -> lowerState_, th_, phi_, opts : OptionsPattern[]] /; AtomicSystemType[sys] == "Toy" :=
    1 / (4 Pi) Fluorescence[sys, upperState -> lowerState, opts];

unitSphericalVector[th_, phi_] = Simplify@WignerRotate[{{0, 1, 0}}, {phi, th, 0}];

ikk[th_, phi_] = IdentityMatrix[3] - Wedge[ECT@unitSphericalVector[th, phi], unitSphericalVector[th, phi]];

Fluorescence[sys_?AtomicSystemQ, upperState_ -> lowerState_, th_, phi_, opts : OptionsPattern[]] :=
    Tr[FluorescenceTensor[sys, upperState -> lowerState, opts].ikk[th, phi]];

(Fluorescence[sys_?AtomicSystemQ, upperState_ -> lowerState_, opts : OptionsPattern[]] :=
    Module[{subsys, d, b, dm},
        subsys = SelectStates[sys, upperState, lowerState];
        b = energyDifferenceMatrix@subsys branchingRatioMatrix@subsys Boole@DirectedTransitionQ[subsys, upperState -> lowerState];
        d = ComponentMap[b # &, SparseArray@WignerEckart[subsys, {"NormalizedDipole", 1}]];
        dm = DensityMatrix[subsys, FilterRules@opts] degeneracyMatrix@subsys;
        Tr@Vee[dm, d, ECT@d]
    ]) // ProcessOptions;

(FluorescenceTensor[sys_?AtomicSystemQ, upperState_ -> lowerState_, opts : OptionsPattern[]] :=
    Module[{subsys, d, b, dm},
        subsys = SelectStates[sys, upperState, lowerState];
        b = energyDifferenceMatrix@subsys branchingRatioMatrix@subsys Boole@DirectedTransitionQ[subsys, upperState -> lowerState];
        d = ComponentMap[b # &, SparseArray@WignerEckart[subsys, {"NormalizedDipole", 1}]];
        dm = DensityMatrix[subsys, FilterRules@opts] degeneracyMatrix@subsys;
        3 / (8 Pi) TensorTrace[ECT@d, dm, d, Alignment -> Bottom]
    ]) // ProcessOptions;

End[];

EndPackage[];
